@@ -4,44 +4,44 @@ from __future__ import division
 
 import math
 
-from copy import deepcopy
-
-from compas.geometry.basic import scale_vector
-from compas.geometry.basic import scale_vector_xy
-from compas.geometry.basic import normalize_vector
-from compas.geometry.basic import add_vectors
-from compas.geometry.basic import add_vectors_xy
-from compas.geometry.basic import subtract_vectors
-from compas.geometry.basic import subtract_vectors_xy
-from compas.geometry.basic import dot_vectors
-from compas.geometry.basic import cross_vectors
-from compas.geometry.basic import vector_component
-from compas.geometry.basic import vector_component_xy
-from compas.geometry.basic import length_vector
-from compas.geometry.basic import multiply_matrix_vector
-
-from compas.geometry.angles import angle_vectors
-
-from compas.geometry.distance import closest_point_on_plane
-from compas.geometry.distance import closest_point_on_line
-from compas.geometry.distance import closest_point_on_line_xy
-
-from compas.geometry.intersections import intersection_line_plane
-from compas.geometry.intersections import intersection_line_triangle
-
-from compas.geometry.transformations import _EPS
-from compas.geometry.transformations import _SPEC2TUPLE
-from compas.geometry.transformations import _NEXT_SPEC
-
-from compas.geometry.transformations import transform_points
-from compas.geometry.transformations import transform_points_numpy
+from compas.geometry import scale_vector
+from compas.geometry import scale_vector_xy
+from compas.geometry import normalize_vector
+from compas.geometry import add_vectors
+from compas.geometry import add_vectors_xy
+from compas.geometry import subtract_vectors
+from compas.geometry import subtract_vectors_xy
+from compas.geometry import dot_vectors
+from compas.geometry import cross_vectors
+from compas.geometry import vector_component
+from compas.geometry import vector_component_xy
+from compas.geometry import multiply_matrix_vector
+from compas.geometry import multiply_matrices
+from compas.geometry import transpose_matrix
+from compas.geometry import norm_vector
+from compas.geometry import angle_vectors
+from compas.geometry import closest_point_on_plane
+from compas.geometry import closest_point_on_line
+from compas.geometry import closest_point_on_line_xy
+from compas.geometry import intersection_line_plane
+from compas.geometry import intersection_line_triangle
 
 from compas.geometry.transformations import matrix_from_axis_and_angle
 from compas.geometry.transformations import matrix_from_scale_factors
-from compas.geometry.transformations import matrix_from_axis_and_angle
+from compas.geometry.transformations import matrix_from_change_of_basis
 
 
 __all__ = [
+    'local_axes',
+    'orthonormalize_axes',
+
+    'transform_points',
+    'transform_vectors',
+    'transform_frames',
+
+    'local_to_world_coordinates',
+    'world_to_local_coordinates',
+
     'translate_points',
     'translate_points_xy',
 
@@ -51,6 +51,7 @@ __all__ = [
     'rotate_points',
     'rotate_points_xy',
 
+    'mirror_vector_vector',
     'mirror_points_point',
     'mirror_points_point_xy',
     'mirror_points_line',
@@ -60,6 +61,8 @@ __all__ = [
     'project_point_plane',
 
     'project_points_plane',
+    'project_point_line',
+    'project_point_line_xy',
     'project_points_line',
     'project_points_line_xy',
 
@@ -70,6 +73,271 @@ __all__ = [
 ]
 
 
+# this function will not always work
+# it is also a duplicate of stuff found in matrices and frame
+def local_axes(a, b, c):
+    u = b - a
+    v = c - a
+    w = cross_vectors(u, v)
+    v = cross_vectors(w, u)
+    return normalize_vector(u), normalize_vector(v), normalize_vector(w)
+
+
+def orthonormalize_axes(xaxis, yaxis):
+    """Corrects xaxis and yaxis to be unit vectors and orthonormal.
+
+    Parameters
+    ----------
+    xaxis: :class:`Vector` or list of float
+    yaxis: :class:`Vector` or list of float
+
+    Returns
+    -------
+    tuple: (xaxis, yaxis)
+        The corrected axes.
+
+    Raises
+    ------
+    ValueError: If xaxis and yaxis cannot span a plane.
+
+    Examples
+    --------
+    >>> xaxis = [1, 4, 5]
+    >>> yaxis = [1, 0, -2]
+    >>> xaxis, yaxis = orthonormalize_axes(xaxis, yaxis)
+    >>> allclose(xaxis, [0.1543, 0.6172, 0.7715], tol=0.001)
+    True
+    >>> allclose(yaxis, [0.6929, 0.4891, -0.5298], tol=0.001)
+    True
+    """
+    xaxis = normalize_vector(xaxis)
+    yaxis = normalize_vector(yaxis)
+    zaxis = cross_vectors(xaxis, yaxis)
+    if not norm_vector(zaxis):
+        raise ValueError("Xaxis and yaxis cannot span a plane.")
+    yaxis = cross_vectors(normalize_vector(zaxis), xaxis)
+    return xaxis, yaxis
+
+
+def homogenize(vectors, w=1.0):
+    """Homogenise a list of vectors.
+
+    Parameters
+    ----------
+    vectors : list
+        A list of vectors.
+    w : float, optional
+        Homogenisation parameter.
+        Defaults to ``1.0``.
+
+    Returns
+    -------
+    list
+        Homogenised vectors.
+
+    Notes
+    -----
+    Vectors described by XYZ components are homogenised by appending a homogenisation
+    parameter to the components, and by dividing each component by that parameter.
+    Homogenisatioon of vectors is often used in relation to transformations.
+
+    Examples
+    --------
+    >>> vectors = [[1.0, 0.0, 0.0]]
+    >>> homogenize(vectors)
+    [[1.0, 0.0, 0.0, 1.0]]
+
+    """
+    return [[x * w, y * w, z * w, w] if w else [x, y, z, 0.0] for x, y, z in vectors]
+
+
+def dehomogenize(vectors):
+    """Dehomogenise a list of vectors.
+
+    Parameters
+    ----------
+    vectors : list
+        A list of vectors.
+
+    Returns
+    -------
+    list
+        Dehomogenised vectors.
+
+    Examples
+    --------
+    >>>
+
+    """
+    return [[x / w, y / w, z / w] if w else [x, y, z] for x, y, z, w in vectors]
+
+
+def homogenize_and_flatten_frames(frames):
+    """Homogenize a list of frames and flatten the 3D list into a 2D list.
+
+    Parameters
+    ----------
+    frames: list of :class:`Frame`
+
+    Returns
+    -------
+    list of list of float
+
+    Examples
+    --------
+    >>> frames = [Frame((1, 1, 1), (0, 1, 0), (1, 0, 0))]
+    >>> homogenize_and_flatten_frames(frames)
+    [[1.0, 1.0, 1.0, 1.0], [0.0, 1.0, 0.0, 0.0], [1.0, -0.0, 0.0, 0.0]]
+    """
+    def homogenize_frame(frame):
+        return homogenize([frame[0]], w=1.0) + homogenize([frame[1], frame[2]], w=0.0)
+    return [v for frame in frames for v in homogenize_frame(frame)]
+
+
+def dehomogenize_and_unflatten_frames(points_and_vectors):
+    """Dehomogenize a list of vectors and unflatten the 2D list into a 3D list.
+
+    Parameters
+    ----------
+    points_and_vectors: list of list of float
+        Homogenized points and vectors.
+
+    Returns
+    -------
+    list of list of list of float
+        The frames.
+
+    Examples
+    --------
+    >>> points_and_vectors = [(1., 1., 1., 1.), (0., 1., 0., 0.), (1., 0., 0., 0.)]
+    >>> dehomogenize_and_unflatten_frames(points_and_vectors)
+    [[[1.0, 1.0, 1.0], [0.0, 1.0, 0.0], [1.0, 0.0, 0.0]]]
+    """
+    frames = dehomogenize(points_and_vectors)
+    return [frames[i:i+3] for i in range(0, len(frames), 3)]
+
+
+# ==============================================================================
+# transform
+# ==============================================================================
+
+
+def transform_points(points, T):
+    """Transform multiple points with one transformation matrix.
+
+    Parameters
+    ----------
+    points : list of :class:`Point` or list of list of float
+        A list of points to be transformed.
+    T : :class:`Transformation` or list of list of float
+        The transformation to apply.
+
+    Examples
+    --------
+    >>> points = [[1, 0, 0], [1, 2, 4], [4, 7, 1]]
+    >>> T = matrix_from_axis_and_angle([0, 2, 0], math.radians(45), point=[4, 5, 6])
+    >>> points_transformed = transform_points(points, T)
+    """
+    return dehomogenize(multiply_matrices(homogenize(points, w=1.0), transpose_matrix(T)))
+
+
+def transform_vectors(vectors, T):
+    """Transform multiple vectors with one transformation matrix.
+
+    Parameters
+    ----------
+    vectors : list of :class:`Vector` or list of list of float
+        A list of vectors to be transformed.
+    T : :class:`Transformation` list of list of float
+        The transformation to apply.
+
+    Examples
+    --------
+    >>> vectors = [[1, 0, 0], [1, 2, 4], [4, 7, 1]]
+    >>> T = matrix_from_axis_and_angle([0, 2, 0], math.radians(45), point=[4, 5, 6])
+    >>> vectors_transformed = transform_vectors(vectors, T)
+    """
+    return dehomogenize(multiply_matrices(homogenize(vectors, w=0.0), transpose_matrix(T)))
+
+
+def transform_frames(frames, T):
+    """Transform multiple frames with one transformation matrix.
+
+    Parameters
+    ----------
+    frames : list of :class:`Frame`
+        A list of frames to be transformed.
+    T : :class:`Transformation`
+        The transformation to apply on the frames.
+
+    Examples
+    --------
+    >>> frames = [Frame([1, 0, 0], [1, 2, 4], [4, 7, 1]), Frame([0, 2, 0], [5, 2, 1], [0, 2, 1])]
+    >>> T = matrix_from_axis_and_angle([0, 2, 0], math.radians(45), point=[4, 5, 6])
+    >>> transformed_frames = transform_frames(frames, T)
+    """
+    points_and_vectors = homogenize_and_flatten_frames(frames)
+    return dehomogenize_and_unflatten_frames(multiply_matrices(points_and_vectors, transpose_matrix(T)))
+
+
+def world_to_local_coordinates(frame, xyz):
+    """Convert global coordinates to local coordinates.
+
+    Parameters
+    ----------
+    frame : :class:`Frame` or [point, xaxis, yaxis]
+        The local coordinate system.
+    xyz : array-like
+        The global coordinates of the points to convert.
+
+    Returns
+    -------
+    list of list of float
+        The coordinates of the given points in the local coordinate system.
+
+
+    Examples
+    --------
+    >>> import numpy as np
+    >>> f = Frame([0, 1, 0], [3, 4, 1], [1, 5, 9])
+    >>> xyz = [Point(2, 3, 5)]
+    >>> Point(*world_to_local_coordinates(f, xyz)[0])
+    Point(3.726, 4.088, 1.550)
+    """
+    from compas.geometry import Frame  # noqa: F811
+    T = matrix_from_change_of_basis(Frame.worldXY(), frame)
+    return transform_points(xyz, T)
+
+
+def local_to_world_coordinates(frame, xyz):
+    """Convert local coordinates to global coordinates.
+
+    Parameters
+    ----------
+    frame : :class:`Frame` or [point, xaxis, yaxis]
+        The local coordinate system.
+    xyz : list of `Points` or list of list of float
+        The global coordinates of the points to convert.
+
+    Returns
+    -------
+    list of list of float
+        The coordinates of the given points in the local coordinate system.
+
+
+    Examples
+    --------
+    >>> import numpy as np
+    >>> f = Frame([0, 1, 0], [3, 4, 1], [1, 5, 9])
+    >>> xyz = [Point(3.726, 4.088, 1.550)]
+    >>> Point(*local_to_world_coordinates(f, xyz)[0])
+    Point(2.000, 3.000, 5.000)
+    """
+    from compas.geometry import Frame  # noqa: F811
+    T = matrix_from_change_of_basis(frame, Frame.worldXY())
+    return transform_points(xyz, T)
+
+
 # ==============================================================================
 # translate
 # ==============================================================================
@@ -92,7 +360,7 @@ def translate_points(points, vector):
 
     Examples
     --------
-    >>> 
+    >>>
 
     """
     return [add_vectors(point, vector) for point in points]
@@ -115,7 +383,7 @@ def translate_points_xy(points, vector):
 
     Examples
     --------
-    >>> 
+    >>>
 
     """
     return [add_vectors_xy(point, vector) for point in points]
@@ -143,7 +411,7 @@ def scale_points(points, scale):
 
     Examples
     --------
-    >>> 
+    >>>
 
     """
     T = matrix_from_scale_factors([scale, scale, scale])
@@ -167,7 +435,7 @@ def scale_points_xy(points, scale):
 
     Examples
     --------
-    >>> 
+    >>>
 
     """
     T = matrix_from_scale_factors([scale, scale, 0])
@@ -428,10 +696,10 @@ def mirror_points_line(points, line):
         The mirrored points.
 
     """
-    return [closest_point_on_line(point, line) for point in points]
+    return [mirror_point_line(point, line) for point in points]
 
 
-def mirror_points_line_xy(point, line):
+def mirror_points_line_xy(points, line):
     """Mirror a point about a line.
 
     Parameters
@@ -447,7 +715,7 @@ def mirror_points_line_xy(point, line):
         The mirrored points.
 
     """
-    return [closest_point_on_line_xy(point, line) for point in points]
+    return [mirror_point_line_xy(point, line) for point in points]
 
 
 def mirror_point_plane(point, plane):
@@ -528,7 +796,7 @@ def project_point_plane(point, plane):
     >>> point = [3.0, 3.0, 3.0]
     >>> plane = ([0.0, 0.0, 0.0], [0.0, 0.0, 1.0])  # the XY plane
     >>> project_point_plane(point, plane)
-    [3.0, 3.0, 3.0]
+    [3.0, 3.0, 0.0]
 
     """
     base, normal = plane
@@ -770,7 +1038,7 @@ def reflect_line_triangle(line, triangle, tol=1e-6):
     >>> triangle = [1.0, 0, 0], [-1.0, 0, 0], [0, 0, 1.0]
     >>> line = [-1, 1, 0], [-0.5, 0.5, 0]
     >>> reflect_line_triangle(line, triangle)
-    ([0.0, 0.0, 0], [1.0, 1.0, 0])
+    ([0.0, 0.0, 0.0], [1.0, 1.0, 0.0])
 
     """
     x = intersection_line_triangle(line, triangle, tol=tol)
@@ -824,32 +1092,29 @@ def orient_points(points, reference_plane, target_plane):
 
     Examples
     --------
-    .. code-block:: python
-
-        from compas.geometry import orient_points
-        from compas.geometry import intersection_segment_segment_xy
-
-        refplane = ([0.57735, 0.57735, 0.57735], [1.0, 1.0, 1.0])
-        tarplane = ([0.0, 0.0, 0.0], [0.0, 0.0, 1.0])
-
-        points = [
-            [0.288675, 0.288675, 1.1547],
-            [0.866025, 0.866025, 0.0],
-            [1.077350, 0.077350, 0.57735],
-            [0.077350, 1.077350, 0.57735]
+    >>> from compas.geometry import orient_points
+    >>> from compas.geometry import intersection_segment_segment_xy
+    >>>
+    >>> refplane = ([0.57735, 0.57735, 0.57735], [1.0, 1.0, 1.0])
+    >>> tarplane = ([0.0, 0.0, 0.0], [0.0, 0.0, 1.0])
+    >>>
+    >>> points = [\
+            [0.288675, 0.288675, 1.1547],\
+            [0.866025, 0.866025, 0.0],\
+            [1.077350, 0.077350, 0.57735],\
+            [0.077350, 1.077350, 0.57735]\
         ]
-
-        points = orient_points(points, refplane, tarplane)
-
-        ab = points[0], points[1]
-        cd = points[2], points[3]
-
-        point = intersection_segment_segment_xy(ab, cd)
-
-        points = orient_points([point], tarplane, refplane)
-        
-        print(points[0])
-
+    >>>
+    >>> points = orient_points(points, refplane, tarplane)
+    >>>
+    >>> ab = points[0], points[1]
+    >>> cd = points[2], points[3]
+    >>>
+    >>> point = intersection_segment_segment_xy(ab, cd)
+    >>>
+    >>> points = orient_points([point], tarplane, refplane)
+    >>> Point(*points[0])
+    Point(0.577, 0.577, 0.577)
     """
     axis = cross_vectors(reference_plane[1], target_plane[1])
     angle = angle_vectors(reference_plane[1], target_plane[1])
@@ -870,42 +1135,10 @@ def orient_points(points, reference_plane, target_plane):
 
 if __name__ == "__main__":
 
-    from compas.geometry import orient_points
-    from compas.geometry import intersection_segment_segment_xy
+    import doctest
 
+    from compas.geometry import allclose  # noqa: F401
+    from compas.geometry import Point  # noqa: F401
+    from compas.geometry import Frame  # noqa: F401
 
-    refplane = ([0.57735, 0.57735, 0.57735], [1.0, 1.0, 1.0])
-    tarplane = ([0.0, 0.0, 0.0], [0.0, 0.0, 1.0])
-
-    points = [
-        [0.288675, 0.288675, 1.1547],
-        [0.866025, 0.866025, 0.0],
-        [1.077350, 0.077350, 0.57735],
-        [0.077350, 1.077350, 0.57735]
-    ]
-
-    points = orient_points(points, refplane, tarplane)
-
-    ab = points[0], points[1]
-    cd = points[2], points[3]
-
-    point = intersection_segment_segment_xy(ab, cd)
-
-    points = orient_points([point], tarplane, refplane)
-    
-    print(points[0])
-
-    # points = [
-    #     [ 1.0,  1.0, 0.0],
-    #     [-1.0,  1.0, 0.0],
-    #     [-1.0, -1.0, 0.0],
-    #     [ 1.0, -1.0, 0.0]
-    # ]
-
-    # refplane = ([0, 0, 0], [0, 0, -1.0])
-    # tarplane = ([0, 0, 0], [0, 0, 1.0])
-
-    # points = orient_points(points, refplane, tarplane)
-
-    # print(points)
-
+    doctest.testmod(globs=globals())
